<!doctype html>
<html>
    <head>
        <title>Hello World</title>
    </head>

    <body>
        <p>Hello {{name}}</p>
        <form action='/backtest' method='POST'>
            <input type="date" name="start">
            <input type="date" name="end">
            <input type="number" name='capital' value=100000>
            <select name="universe">
                {% for u in universe %}
                    <option value="{{u}}">{{u}}</option>
                {% endfor %}
            </select>
            <select name="strategy">
                {% for s in strategies %}
                    <option value="{{s}}">{{s}}</option>
                {% endfor %}
            </select>
            <input type="decimal" name="stop_loss" placeholder="stop_loss">
            <input type="submit" value="Submit">
        </form>
    </body>
</html>
